"""
Topic -- organizes cataloged Zope objects.
"""

# TODO:
#

import time, copy
import Products
from intSet import intSet
from OFS.SimpleItem import Item
from OFS.ObjectManager import ObjectManager
from OFS.PropertyManager import PropertyManager
from AccessControl.Role import RoleManager
from webdav.Collection import Collection
from Globals import HTMLFile
from DocumentTemplate.DT_Util import Eval, expr_globals, InstanceDict
from OFS.FindSupport import td

addForm=HTMLFile('topicAdd', globals())

def add(self, id, title='', createPublic=0, createRSS=0, REQUEST=None):
    """
    Add a new ZTopic
    """
    ob=Topic(id, title)
    self._setObject(id, ob)
    try: user=REQUEST['AUTHENTICATED_USER']
    except: user=None
    if createPublic:
        if (user is not None) and not (
            user.has_permission('Add Documents, Images, and Files', self)):
            raise 'Unauthorized', (
                  'You are not authorized to add DTML Documents.'
                  )
        ob.manage_addDTMLMethod(id='index_html', title='', file=defaultTopicView)
    if createRSS:
        if (user is not None) and not (
            user.has_permission('Add Documents, Images, and Files', self)):
            raise 'Unauthorized', (
                  'You are not authorized to add DTML Documents.'
                  )
        ob.manage_addDTMLMethod(id='RSS', title='', file=rssTopicView)    
    if REQUEST is not None:
        return self.manage_main(self, REQUEST, update_menu=1)
        

class Topic(ObjectManager, PropertyManager, RoleManager, Collection, Item):
    """
    Topic -- organizes cataloged Zope objects.
    """
    
    meta_type='Z Topic'

    manage_options=(
        {'label':'Contents', 'action':'manage_main'},
        {'label':'Search Criteria', 'action':'editForm'},
        {'label':'Search Options', 'action':'optionsForm'},
        {'label':'View', 'action':'index_html'},
        {'label':'Properties', 'action':'manage_propertiesForm'},
        {'label':'Import/Export', 'action':'manage_importExportForm'},
        {'label':'Security', 'action':'manage_access'},
        {'label':'Undo', 'action':'manage_UndoForm'},
    )
    
    __ac_permissions__=(
        ('Access contents information', ('contents', 'parentTopics')),
        ('Change Topics', ('editSearches', 'editSearch', 'deleteSearch', 'addSearch')),
        ('View management screens', ('editForm', 'editSearchForm', 'optionsForm')),
        )

    # a list of dictionaries which map search terms to search expressions
    # search expressions are evaluated at search time with the standard Zope
    # expression machinary, and can thus pull from the REQUEST and the Topic
    # object
    criteria=[]

    # boolean, acquire parent's seach criteria, effectively
    # narrowing the search
    acquire_criteria=1
    
    #  boolean, exclude children's search results
    exclude_children=1

    _cached_time=0
    _cached_contentsSet=None
    
    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'description', 'type': 'text', 'mode': 'w'},
    )
    
    editForm=HTMLFile('topicEdit', globals())
    optionsForm=HTMLFile('topicOptions', globals())
    editSearchForm=HTMLFile('searchEdit', globals())
    
    def __init__(self, id, title="", description=""):
        self.id=id
        self.title=title
        self.description=description
    
    def editOptions(self, acquire_criteria=0, 
            exclude_children=0, catalog_id=None, REQUEST=None):
        """
        Edit search criteria
        """
        self._clear_cache()
        if REQUEST is not None and REQUEST.has_key('clear_cache'):
            return self.optionsForm(self, REQUEST, 
                    manage_tabs_message="Cache cleared.")
        self.acquire_criteria=acquire_criteria
        self.exclude_children=exclude_children
	self.catalog_id = catalog_id
        if REQUEST is not None:
            return self.optionsForm(self, REQUEST, 
                    manage_tabs_message="Search options changed.")
    
    def deleteSearch(self, index, REQUEST):
        """
        Delete catalog searches.
        """
        self._clear_cache()
        self.criteria=self.criteria[:index] + self.criteria[index+1:]
        if REQUEST is not None:
            return self.editForm(self, REQUEST,
                    manage_tabs_message="Search deleted.")

    def editSearch(self, index, REQUEST):
        """
        Edit a catalog search.
        """
        self._clear_cache()
        d={}
        for k, v in REQUEST.form.items():
            if k != 'index' and v:
                d[k]=v
                
                # test the expression
                self._eval_expression(v)

        self.criteria=self.criteria[:index] + [d] + self.criteria[index+1:]
        if REQUEST is not None:
            return self.editForm(self, REQUEST,
                    manage_tabs_message="Search changed.")

    def addSearch(self, REQUEST=None):
        """
        Add a search on a catalog
        """
        self._clear_cache()
        self.criteria=self.criteria + [{}]
        if REQUEST is not None:
            return self.editSearchForm(self, REQUEST, index=len(self.criteria)-1)
        
    def contents(self, path=None):
        """
        The objects that match the Topic's seach criteria.
        
        If 'path' is true, then only the paths to the objects
        will be returned, not the actual objects themselves.
        """

	if not hasattr(self, 'catalog_id'):
	    raise ValueError("Catalog ID has not been set")

        r=self._contentsSet()

        if self.acquire_criteria:
            obj=self
            while hasattr(obj, 'aq_parent'):
                obj=obj.aq_parent
                if hasattr(obj, 'aq_base') and obj.aq_base.meta_type=='Z Topic':
		    set = obj._contentsSet()
		    r=r.intersection(set)
                    if not obj.acquire_criteria:
                        break
        
        if self.exclude_children:
            for child in self.objectValues('Z Topic'):
		set = child._contentsSet()
		r=r.intersection(r.difference(set))
        
        # ignore optimization of results for now
        
        result=[]
        if path:
	    c=getattr(self, self.catalog_id)
	    for id in r:
		try:
		    path=c.getpath(id)
		    if path not in result:
			result.append(path)
		except: pass
            return result
        else:
	    c=getattr(self, self.catalog_id)
	    for id in r:
		try:
		    obj=c.getobject(id)
		    if obj not in result:
			result.append(obj)
		except: pass
            return result
            
    def topic_ids(self):
        # avoid acquiring this method
        if self.meta_type=='Z Topic':
            return self.contents(1)

    def _contentsSet(self):
        # contents as intset of results ids

	catalog = getattr(self, self.catalog_id)
	if catalog._catalog.bobobase_modification_time() < self._cached_time:
	    return self._cached_contentsSet
        
        r=intSet()
        for mapping in self.criteria:
            search_map={}
            for k, expr in mapping.items():
                search_map[k]=self._eval_expression(expr)
            results=intSet()
            zcatalog=getattr(self, self.catalog_id)
            c=zcatalog._catalog
            for search_term in search_map.keys():
                if c.indexes.has_key(search_term):
                    index=c.indexes[search_term]
                    search_results=index._apply_index(search_map)
                    if search_results is not None:
                        set, cols=search_results
                        if len(results):
                            results=results.intersection(set)
                        else:
                            results=set

	    if len(r): r = r.union(results)
	    else: r = results
	
	# Make sure r is an intSet - some indexes return other things
	if type(r) != type(intSet()):
	    to_intSet = intSet()
	    map(to_intSet.insert, r.keys())
	    r = to_intSet

        self._cached_contentsSet=r
        self._cached_time=time.time()
        return r
    
    def _clear_cache(self):        
        # avoid unnecessary cache clearing and endless loops
        if self._cached_time==0:
            return
        
        self._cached_time=0
        self._cached_contentsSet=None
        
        # clear child caches if they rely on this Topic
        for obj in self.objectValues('Z Topic'):
            if obj.acquire_criteria:
                obj._clear_cache()
                
        # clear parent caches if they rely on this Topic
        obj=self
        while hasattr(obj, 'aq_parent'):
            obj=obj.aq_parent
            if hasattr(obj, 'aq_base') and obj.aq_base.meta_type=='Z Topic':
                if obj.exclude_children:
                    obj._clear_cache()
                
    def parentTopics(self):
        """
        A list of parent ZTopics. To be called from DTML.
        """
        l=[]
        obj=self
        while hasattr(obj, 'aq_parent'):
            obj=obj.aq_parent
            if hasattr(obj, 'aq_base') and obj.aq_base.meta_type=='Z Topic':
                l.insert(0, obj)
        return l
        
    def _eval_expression(self, expression, REQUEST=None):
        # for evaluating search expressions
        
        if REQUEST is None:
            try: REQUEST=self.REQUEST
            except: pass
        md=td()
        if hasattr(REQUEST, 'AUTHENTICATED_USER'):
            md.AUTHENTICATED_USER=REQUEST.AUTHENTICATED_USER
        expr=Eval(expression, expr_globals)
        md._push(InstanceDict(self, md))
        try: r=expr.eval(md)
        finally:
            md._pop()
            return r

    def Catalogs(self):
	"""Get the list of ZCatalog-like objects available to be Topic'd"""
	folder = self.aq_parent
	catalogs = []
	ZCatalog = Products.ZCatalog.ZCatalog.ZCatalog
	while 1:
	    for item in folder.objectValues() + [folder]:
		if item.__class__ == ZCatalog or \
		   ZCatalog in all_bases(item.__class__):
		    catalogs.append(item)
	    if folder.isTopLevelPrincipiaApplicationObject:
		break
	    folder = folder.aq_parent
	return catalogs

def all_bases(clss):
    bases = list(clss.__bases__)
    for base in bases:
	if base.__bases__:
	    bases = bases + all_bases(base)
    return bases
	    
        
defaultTopicView="""\
<dtml-var standard_html_header>

<p>
<dtml-in parentTopics>
<a href="<dtml-var absolute_url>"><dtml-var id></a>
<dtml-unless sequence-end> / </dtml-unless>
</dtml-in>
</p>

<h2><dtml-var title_or_id></h2>
<p><dtml-var description></p>

<dtml-if "objectValues('Z Topic')">
<h3>Sub-Topics</h3>
<ul>
<dtml-in "objectValues('Z Topic')">
<li><a href="<dtml-var absolute_url>"><dtml-var title_or_id></a></li>
</dtml-in>
</ul>
</dtml-if>

<dtml-if contents>
<h3>Contents</h3>
<ul>
<dtml-in contents skip_unauthorized>
<li><a href="<dtml-var absolute_url>"><dtml-var title_or_id></a></li>
</dtml-in>
</p>
<dtml-else>
<em>This Topic currently has no contents.</em>
</dtml-if>

<dtml-var standard_html_footer>"""


rssTopicView="""\
<!--#call "RESPONSE.setHeader('content-type', 'text/xml')"--><?xml version="1.0"?>

<!DOCTYPE rss PUBLIC "-//Netscape Communications//DTD RSS 0.91//EN"
             "http://my.netscape.com/publish/formats/rss-0.91.dtd">

<rss version="0.91">
<channel>   

<title><dtml-var title_or_id></title>
<link><dtml-var absolute_url></link>
<description><dtml-var description></description>
<language>en-us</language>

<dtml-in contents skip_unauthorized>
<item>
<title><dtml-var title_or_id></title>
<link><dtml-var absolute_url></link>
</item>
</dtml-in>

</channel>
</rss>"""
